# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import hashlib

from django.db import models, IntegrityError
from django.contrib.auth import get_user_model
from django_redis import get_redis_connection
from django.utils.translation import ugettext as _
from django.utils.encoding import force_text

from .settings import INFO_PREFIX, INFO_TIMEOUT, PREFIX
from .session import SessionStore, DataBaseStore
from .constants import MAX_KEY_LENGTH, MAX_PREFIX_LENGTH, GLOBAL_PREFIX, INFO_KEY, DEFAULT_PREFIX, INFO_EXIST_VALUE


class SessionInfo(models.Model):
    """
    Session-related information
    """
    user = models.ForeignKey(get_user_model(), related_name='sessions', db_index=True, verbose_name=_('User'))
    creation_date = models.DateTimeField(auto_now_add=True, verbose_name=_('Creation date'))
    key = models.CharField(max_length=MAX_KEY_LENGTH, db_index=True, verbose_name=_('Key'))
    prefix = models.CharField(max_length=MAX_PREFIX_LENGTH, db_index=True, default=DEFAULT_PREFIX)
    active = models.BooleanField(default=True)

    # additional request data
    user_ip = models.IPAddressField(default='127.0.0.1')
    user_agent = models.TextField(null=True, default=None, blank=True)

    # auto-generated data on .save()
    user_agent_md5 = models.CharField(max_length=32, default=None, null=True, blank=True)

    @classmethod
    def get_key(cls, session_key):
        return ':'.join([INFO_PREFIX, GLOBAL_PREFIX, INFO_KEY, session_key])

    @classmethod
    def key_exists(cls, session_key):
        """
        If session info exists with that session_key, returns True
        """
        key = cls.get_key(session_key)
        conn = get_redis_connection()
        return conn.exists(key)

    @staticmethod
    def exists(request):
        return SessionInfo.key_exists(request.session.session_key)

    @classmethod
    def gen(cls, request):
        """
        Generate Session model from request
        """
        session = SessionInfo(user=request.user, key=request.session.session_key)

        # additional request data processing
        session.user_ip = request.META.get('REMOTE_ADDR')
        session.user_agent = force_text(request.META.get('HTTP_USER_AGENT'), errors='replace')

        return session

    @classmethod
    def record(cls, request, prefix=PREFIX):
        """
        Record session info from request
        """
        session_key = request.session.session_key
        try:
            session = cls.gen(request)
            session.prefix = prefix
            session.save()
        except IntegrityError:
            # session info is already in database
            pass
        conn = get_redis_connection()
        conn.setex(cls.get_key(session_key), INFO_TIMEOUT, INFO_EXIST_VALUE)

    def erase(self):
        self.active = False
        SessionStore(self.key, self.prefix).delete()
        self.save()

    def session_exists(self):
        return SessionStore(self.key, self.prefix).exists()
    session_exists.boolean = True

    def session_exists_db(self):
        return DataBaseStore().exists(self.key)
    session_exists_db.boolean = True

    def session_data(self):
        s = SessionStore(self.key, self.prefix)
        if s.exists(self.key):
            return s.load()
        if self.session_exists_db():
            return DataBaseStore(self.key).load()
        return _('Does not exist')

    def __unicode__(self):
        return '{user}'.format(user=self.user)

    def save(self, *args, **kwargs):
        if self.user_agent:
            self.user_agent_md5 = hashlib.md5(self.user_agent).hexdigest()
        super(SessionInfo, self).save(*args, **kwargs)

    class Meta:
        verbose_name = _('Session')
        verbose_name_plural = _('Sessions')
        unique_together = (('prefix', 'key'),)


# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import logging

from .models import SessionInfo

logger = logging.getLogger(__name__)


class SessionSyncMiddleware(object):
    @staticmethod
    def process_response(request, response):
        if hasattr(request, 'user') and request.user.is_authenticated() and not SessionInfo.exists(request):
            SessionInfo.record(request)
        return response


# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured

from .constants import MAX_PREFIX_LENGTH

SETTINGS_PREFIX = 'DD_SESSION'


def get(key, default):
    return getattr(settings, '_'.join([SETTINGS_PREFIX, key]), default)

PREFIX = get('PREFIX', 'default')                      # prefix for data keys
INFO_TIMEOUT = get('INFO_TIMEOUT', 60 * 60 * 24)       # timeout of keeping info existence key
INFO_PREFIX = get('INFO_PREFIX', PREFIX)               # prefix for info existence keys
USE_FALLBACK = get('USE_FALLBACK', True)               # use read-only fallback to database sessions

if len(PREFIX) > MAX_PREFIX_LENGTH:
    raise ImproperlyConfigured('Prefix %s is more that MAX_PREFIX_LENGTH %s' % (PREFIX, MAX_PREFIX_LENGTH))


# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import logging
from redis.exceptions import RedisError

from django.utils.encoding import force_unicode
from django.contrib.sessions.backends.base import SessionBase, CreateError
from django.core.exceptions import SuspiciousOperation
from django.contrib.sessions.backends.db import Session as DataBaseSession, SessionStore as DataBaseStore
from django_redis import get_redis_connection
from django.utils import timezone


from .settings import PREFIX, USE_FALLBACK
from .constants import DATA_KEY, GLOBAL_PREFIX

logger = logging.getLogger(__name__)


class SessionStore(SessionBase):
    """
    Implements Redis session store.
    Forked from: https://github.com/martinrusev/django-redis-sessions
    """

    @classmethod
    def clear_expired(cls):
        # clear_expired is no-op and is done by redis
        if USE_FALLBACK:
            logger.info('clearing expired DB sessions')
            DataBaseStore.clear_expired()
        return True

    def __init__(self, session_key=None, prefix=PREFIX):
        super(SessionStore, self).__init__(session_key)
        self.server = get_redis_connection()
        self.prefix = prefix

    def load_db(self):
        try:
            s = DataBaseSession.objects.get(
                session_key=self.session_key,
                expire_date__gt=timezone.now()
            )
            logger.info('fallback successful')
            # noinspection PyAttributeOutsideInit
            self._session_cache = self.decode(s.session_data)
            self.save()
            return self._session_cache
        except (DataBaseSession.DoesNotExist, SuspiciousOperation) as e:
            logger.info('fallback failed: %s' % e)
            self.create()
            return {}

    def load_redis(self):
        logger.info('loading session from redis: %s' % self.session_key)
        try:
            session_data = self.server.get(
                self.redis_key(self.session_key)
            )
            return self.decode(force_unicode(session_data))
        except RedisError as e:
            logger.error('error while loading session %s: %s' % (self.session_key, e))
            self.create()
            return {}

    def load(self):
        if USE_FALLBACK and not self.exists(self.session_key):
            return self.load_db()
        return self.load_redis()

    def exists(self, session_key=None):
        if session_key is None:
            session_key = self.session_key
        return self.server.exists(self.get_redis_key(session_key, self.prefix))

    def create(self):
        logger.info('creating session')
        while True:
            self._session_key = self._get_new_session_key()
            try:
                self.save(must_create=True)
            except CreateError:
                continue
            self.modified = True
            return

    def save(self, must_create=False):
        logger.info('saving session')
        if must_create and self.exists(self._get_or_create_session_key()):
            raise CreateError
        data = self.encode(self._get_session(no_load=must_create))
        self.server.setex(
            self.get_redis_key(self._get_or_create_session_key(), self.prefix),
            self.get_expiry_age(),
            data
        )

    def delete(self, session_key=None):
        if session_key is None:
            if self.session_key is None:
                return
            session_key = self.session_key
        try:
            self.delete_session(session_key, conn=self.server, prefix=self.prefix)
            from .models import SessionInfo
            SessionInfo.objects.filter(key=session_key, prefix=self.prefix).update(active=False)
        except Exception as e:
            logger.warning(e)

    @classmethod
    def delete_session(cls, session_key, conn=None, prefix=PREFIX):
        logger.info('deleting session %s' % session_key)
        if USE_FALLBACK:
            DataBaseStore(session_key).delete()
        if conn is None:
            conn = get_redis_connection()
        return conn.delete(cls.get_redis_key(session_key, prefix))

    @staticmethod
    def get_redis_key(session_key, prefix=PREFIX):
        """Return the real key name in redis storage
        @return string
        """
        return ':'.join([prefix, GLOBAL_PREFIX, DATA_KEY, session_key])

    def redis_key(self, session_key=None):
        if session_key is None:
            session_key = self.session_key
        return self.get_redis_key(session_key, self.prefix)
